"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const css_tree_1 = __importDefault(require("css-tree"));
const url_1 = require("url");
const _1 = require(".");
const data_1 = require("./data");
exports.cssURLPattern = /([:,\s]\s*url)\s*\((?!\s*data:.+,)\s*(\S+?)\s*\)/gi;
const fontFormats = {
    woff2: 'woff2',
    woff: 'woff',
    ttf: 'truetype',
    otf: 'opentype',
};
const fontFormatNames = Object.values(fontFormats);
function extractFontsFromValue(value) {
    const chunks = value.children.toArray().reduce((array, node) => {
        if (node.type === 'Operator' && node.value === ',') {
            return [...array, []];
        }
        else {
            array[array.length - 1].push(node);
            return array;
        }
    }, [[]]);
    const local = [];
    const remote = [];
    chunks.forEach(chunk => {
        const functionNodes = chunk.filter((node) => node.type === 'Function');
        const localNode = functionNodes.find(node => node.name === 'local');
        const formatNode = functionNodes.find(node => node.name === 'format');
        const urlNode = chunk.find((node) => node.type === 'Url' && !data_1.dataURLPattern.test(node.value.value));
        if (localNode) {
            local.push(localNode);
            return;
        }
        let formatName;
        if (formatNode) {
            const data = formatNode.children.first();
            if (data && data.type === 'String') {
                const _formatName = data.value.replace(/['"]/g, '');
                if (fontFormatNames.includes(_formatName)) {
                    formatName = _formatName;
                }
            }
        }
        else if (urlNode) {
            const ext = _1.extractExtensionFromURL(urlNode.value.value);
            if (ext && ext in fontFormats) {
                formatName = fontFormats[ext];
            }
        }
        if (urlNode && formatName) {
            remote.push({ path: urlNode.value.value, formatName });
        }
    });
    return { local, remote };
}
function optimizeCSS(text, currentURL) {
    const doubleQuote = '"';
    const singleQuote = "'";
    const urls = new Set();
    const ast = css_tree_1.default.parse(text);
    css_tree_1.default.walk(ast, node => {
        if (node.type === 'Url') {
            const { value } = node.value;
            const relativeURL = value.startsWith(doubleQuote) && value.endsWith(doubleQuote)
                ? value.slice(1, -1).replace(/\\"/g, doubleQuote)
                : value.startsWith(singleQuote) && value.endsWith(singleQuote)
                    ? value.slice(1, -1).replace(/\\'/g, singleQuote)
                    : value;
            const absoluteURL = url_1.resolve(currentURL, relativeURL);
            node.value.value = absoluteURL;
        }
    });
    css_tree_1.default.walk(ast, node => {
        if (node.type === 'Url') {
            const { value } = node.value;
            if (!data_1.dataURLPattern.test(value)) {
                urls.add(value);
            }
            return;
        }
        if (node.type === 'Atrule' && node.name === 'font-face' && node.block) {
            const { base, local, remote } = node.block.children.toArray().reduce(({ base, local, remote }, line) => {
                if (line.type === 'Declaration' &&
                    line.property === 'src' &&
                    line.value.type === 'Value') {
                    const fonts = extractFontsFromValue(line.value);
                    return {
                        base,
                        local: [...local, ...fonts.local],
                        remote: [...remote, ...fonts.remote],
                    };
                }
                else {
                    return { base: [...base, line], local, remote };
                }
            }, {
                base: [],
                local: [],
                remote: [],
            });
            const optimalRemote = remote.sort((a, b) => {
                return ((fontFormatNames.indexOf(a.formatName) + 1 || 9) -
                    (fontFormatNames.indexOf(b.formatName) + 1 || 9));
            })[0];
            const optimalRemoteNodes = optimalRemote
                ? [
                    {
                        type: 'Url',
                        value: { type: 'Raw', value: optimalRemote.path },
                    },
                    { type: 'WhiteSpace', value: ' ' },
                    {
                        type: 'Function',
                        name: 'format',
                        children: new css_tree_1.default.List().fromArray([
                            { type: 'String', value: `'${optimalRemote.formatName}'` },
                        ]),
                    },
                ]
                : [];
            const chunks = [...local.map(l => [l]), optimalRemoteNodes];
            const children = chunks.reduce((list, chunk) => {
                if (list.getSize())
                    list.appendData({ type: 'Operator', value: ',' });
                chunk.forEach(node => list.appendData(node));
                return list;
            }, new css_tree_1.default.List());
            node.block.children.fromArray([
                {
                    type: 'Declaration',
                    property: 'src',
                    important: false,
                    value: { type: 'Value', children },
                },
                ...base,
            ]);
        }
    });
    return { text: css_tree_1.default.generate(ast), urls };
}
exports.optimizeCSS = optimizeCSS;
//# sourceMappingURL=css.js.map